"""Tests for HTMLExporter"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import re

from nbformat import v4
from traitlets.config import Config

from nbconvert.exporters.html import HTMLExporter

from .base import ExportersTestsBase


class TestHTMLExporter(ExportersTestsBase):
    """Tests for HTMLExporter"""

    exporter_class = HTMLExporter  # type:ignore
    should_include_raw = ["html"]  # type:ignore

    def test_constructor(self):
        """
        Can a HTMLExporter be constructed?
        """
        HTMLExporter()

    def test_export(self):
        """
        Can a HTMLExporter export something?
        """
        (output, resources) = HTMLExporter().from_filename(self._get_notebook())
        assert len(output) > 0

    def test_export_classic(self):
        """
        Can a HTMLExporter export using the 'classic' template?
        """
        (output, resources) = HTMLExporter(template_name="classic").from_filename(
            self._get_notebook()
        )
        assert len(output) > 0

    def test_export_notebook(self):
        """
        Can a HTMLExporter export using the 'lab' template?
        """
        (output, resources) = HTMLExporter(template_name="lab").from_filename(self._get_notebook())
        assert len(output) > 0

    def test_prompt_number(self):
        """
        Does HTMLExporter properly format input and output prompts?
        """
        no_prompt_conf = Config(
            {
                "TemplateExporter": {
                    "exclude_input_prompt": True,
                    "exclude_output_prompt": True,
                }
            }
        )
        exporter = HTMLExporter(config=no_prompt_conf, template_name="lab")
        (output, resources) = exporter.from_filename(
            self._get_notebook(nb_name="prompt_numbers.ipynb")
        )
        in_regex = r"In&nbsp;\[(.*)\]:"
        out_regex = r"Out\[(.*)\]:"

        assert not re.findall(in_regex, output)
        assert not re.findall(out_regex, output)

    def test_png_metadata(self):
        """
        Does HTMLExporter with the 'classic' template treat pngs with width/height metadata correctly?
        """
        exporter = HTMLExporter(template_name="classic")
        with self.assertLogs(exporter.log, level="WARN") as log:
            (output, resources) = exporter.from_filename(
                self._get_notebook(nb_name="pngmetadata.ipynb")
            )
            assert len(log.output) == 1
            assert "Alternative text is missing on 1 image(s)." in log.output[0]

        check_for_png = re.compile(
            r'<img alt="No description has been provided for this image"([^>]*?)>'
        )
        result = check_for_png.search(output)
        assert result
        attr_string = result.group(1)
        assert "width=" in attr_string
        assert "height=" in attr_string

    def test_javascript_output(self):
        nb = v4.new_notebook(
            cells=[
                v4.new_code_cell(
                    outputs=[
                        v4.new_output(
                            output_type="display_data",
                            data={"application/javascript": "javascript_output();"},
                        )
                    ]
                )
            ]
        )
        (output, resources) = HTMLExporter(template_name="classic").from_notebook_node(nb)
        self.assertIn("javascript_output", output)

    def test_mermaid_output(self):
        nb = v4.new_notebook(
            cells=[
                v4.new_code_cell(
                    outputs=[
                        v4.new_output(
                            output_type="display_data",
                            data={"text/vnd.mermaid": "flowchart LR\na --> b"},
                        )
                    ]
                )
            ]
        )
        (output, resources) = HTMLExporter(template_name="lab").from_notebook_node(nb)
        self.assertIn("""<div class="jp-Mermaid">""", output)
        self.assertIn("""<pre class="mermaid">""", output)

    def test_mermaid_prerendered_output(self):
        nb = v4.new_notebook(
            cells=[
                v4.new_code_cell(
                    outputs=[
                        v4.new_output(
                            output_type="display_data",
                            data={
                                "image/svg+xml": "<svg></svg>",
                                "text/vnd.mermaid": "flowchart LR\na --> b",
                            },
                        )
                    ]
                )
            ]
        )
        (output, resources) = HTMLExporter(template_name="lab").from_notebook_node(nb)
        self.assertNotIn("""<div class="jp-Mermaid">""", output)

    def test_attachments(self):
        (output, resources) = HTMLExporter(template_name="classic").from_file(
            self._get_notebook(nb_name="attachment.ipynb")
        )
        check_for_png = re.compile(r'<img alt="image.png" src="([^"]*?)"/>')
        result = check_for_png.search(output)
        assert result
        self.assertTrue(result.group(1).strip().startswith("data:image/png;base64,iVBOR"))

        check_for_data = re.compile(r'<img alt="image.png" src="(?P<url>[^"]*?)"')
        results = check_for_data.findall(output)
        assert results[0] != results[1], "attachments only need to be unique within a cell"
        assert "image/svg" in results[1], "second image should use svg"

    def test_custom_filter_highlight_code(self):
        # Overwriting filters takes place at: Exporter.from_notebook_node
        nb = v4.new_notebook()
        nb.cells.append(v4.new_code_cell("some_text"))

        def custom_highlight_code(source, language="python", metadata=None):
            return source + " ADDED_TEXT"

        filters = {"highlight_code": custom_highlight_code}
        (output, resources) = HTMLExporter(
            template_name="classic", filters=filters
        ).from_notebook_node(nb)
        self.assertTrue("ADDED_TEXT" in output)

    def test_basic_name(self):
        """
        Can a HTMLExporter export using the 'basic' template?
        """
        (output, resources) = HTMLExporter(template_name="basic").from_filename(
            self._get_notebook()
        )
        assert len(output) > 0

    def test_javascript_injection(self):
        for template in ["lab", "classic", "reveal"]:
            (output, resources) = HTMLExporter(template_name=template).from_filename(
                self._get_notebook("notebook_inject.ipynb")
            )

            # Check injection in the metadata.title of the Notebook
            assert "<script>alert('title')</script>" not in output

            # Check injection in the metadata.widgets of the Notebook
            assert "</script><script>alert('widgets')" not in output

            # Check injection in the cell.metadata.tags of the Notebook
            assert "<script>alert('cell_tag')</script>" not in output

            # Check injection in the cell.source of the Notebook
            assert "<script>alert('raw cell')</script>" not in output

            # Check injection in svg output
            assert "<script>alert('image/svg+xml output')</script>" not in output
            assert "<script>alert('svg_filename')</script>" not in output

            # Check injection in image filenames
            assert "<script>alert('png filenames')</script>" not in output
            assert "<script>alert('jpg filenames')</script>" not in output

            # Check injection in image data
            assert "<script>alert('image/png output')</script>" not in output
            assert "<script>alert('image/jpeg output')</script>" not in output

            # Check injection in image width/height
            assert "<script>alert('output.metadata.width png injection')</script>" not in output
            assert "<script>alert('output.metadata.height png injection')</script>" not in output

            # Check injection in widget view
            assert (
                "<script>alert('output.data.application/vnd.jupyter.widget-view+json injection')"
                not in output
            )

        # By design, text/html, text/markdown, application/javascript and markdown cells should allow
        # for JavaScript code execution
        for template in ["lab", "classic", "reveal"]:
            (output, resources) = HTMLExporter(template_name=template).from_filename(
                self._get_notebook("notebook_inject.ipynb")
            )

            assert "<script>alert('markdown cell')</script>" in output
            assert "<script>alert('text/markdown output')</script>" in output
            assert "<script>alert('text/html output')</script>" in output
            assert "alert('application/javascript output')" in output

        # But it's an opt-out
        for template in ["lab", "classic", "reveal"]:
            (output, resources) = HTMLExporter(
                template_name=template, sanitize_html=True
            ).from_filename(self._get_notebook("notebook_inject.ipynb"))

            assert "<script>alert('markdown cell')</script>" not in output
            assert "<script>alert('text/markdown output')</script>" not in output
            assert "<script>alert('text/html output')</script>" not in output
            assert "alert('application/javascript output')" not in output

    def test_language_code_not_set(self):
        (output, resources) = HTMLExporter(template_name="classic").from_filename(
            self._get_notebook()
        )
        assert '<html lang="en">' in output

    def test_set_language_code(self):
        exporter = HTMLExporter(template_name="classic", language_code="fr")
        (output, resources) = exporter.from_filename(self._get_notebook())
        assert '<html lang="fr">' in output

    def test_language_code_error(self):
        with self.assertLogs(level="WARN") as log:
            exporter = HTMLExporter(template_name="classic", language_code="zz")
            assert len(log.output) == 1
            assert '"zz" is not an ISO 639-1 language code.' in log.output[0]
        (output, resources) = exporter.from_filename(self._get_notebook())

        assert '<html lang="en">' in output



"""
Module with tests for exporter.py
"""

# -----------------------------------------------------------------------------
# Copyright (c) 2013, the IPython Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import os
from unittest.mock import patch

from traitlets.config import Config

from nbconvert.exporters import Exporter, TemplateExporter, get_export_names
from nbconvert.preprocessors.base import Preprocessor

from .base import ExportersTestsBase

# -----------------------------------------------------------------------------
# Class
# -----------------------------------------------------------------------------


class PizzaPreprocessor(Preprocessor):
    """Simple preprocessor that adds a 'pizza' entry to the Notebook metadata.  Used
    to test Exporter.
    """

    def preprocess(self, nb, resources):
        nb["metadata"]["pizza"] = "cheese"
        return nb, resources


class DummyExporter(TemplateExporter):
    """
    Dummy exporter to check that parent default_config gets overwritten properly
    """

    @property
    def default_config(self):
        c = Config({"TagRemovePreprocessor": {"enabled": False}})
        if super().default_config:
            c2 = super().default_config.copy()
            c2.merge(c)
            c = c2
        return c


class TestExporter(ExportersTestsBase):
    """Contains test functions for exporter.py"""

    def test_constructor(self):
        """Can an Exporter be constructed?"""
        Exporter()

    def test_export(self):
        """Can an Exporter export something?"""
        exporter = Exporter()
        (notebook, resources) = exporter.from_filename(self._get_notebook())
        assert isinstance(notebook, dict)

    def test_preprocessor(self):
        """Do preprocessors work?"""
        config = Config({"Exporter": {"preprocessors": [PizzaPreprocessor()]}})
        exporter = Exporter(config=config)
        (notebook, resources) = exporter.from_filename(self._get_notebook())
        self.assertEqual(notebook["metadata"]["pizza"], "cheese")

    def test_get_export_names_disable(self):
        """Can we disable all exporters then enable a single one"""
        config = Config({"Exporter": {"enabled": False}, "NotebookExporter": {"enabled": True}})
        export_names = get_export_names(config=config)
        self.assertEqual(export_names, ["notebook"])

    def test_get_exporter_disable_config_exporters(self):
        """
        Does get_export_names behave correctly with respect to
        NBCONVERT_DISABLE_CONFIG_EXPORTERS being set in the
        environment?
        """
        config = Config({"Exporter": {"enabled": False}, "NotebookExporter": {"enabled": True}})
        os.environ["NBCONVERT_DISABLE_CONFIG_EXPORTERS"] = "1"
        with patch("nbconvert.exporters.base.get_exporter") as exp:
            export_names = get_export_names(config=config)
            # get_export_names should not call get_exporter for
            # any of the entry points because we return before then.
            exp.assert_not_called()

            # We should have all exporters, not just the ones
            # enabled in the config
            self.assertNotEqual(export_names, ["notebook"])

        # In the absence of this variable we should revert to
        # the normal behavior.
        del os.environ["NBCONVERT_DISABLE_CONFIG_EXPORTERS"]
        export_names = get_export_names(config=config)
        self.assertEqual(export_names, ["notebook"])

    def test_default_config_merge(self):
        """
        Do default_configs merge properly?
        Class config should overwrite parent config
        """
        e = DummyExporter()
        self.assertFalse(e.default_config["TagRemovePreprocessor"]["enabled"])
        self.assertTrue(e.default_config["RegexRemovePreprocessor"]["enabled"])



"""reStructuredText Exporter class"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from traitlets import default
from traitlets.config import Config

from .templateexporter import TemplateExporter


class RSTExporter(TemplateExporter):
    """
    Exports reStructuredText documents.
    """

    @default("file_extension")
    def _file_extension_default(self):
        return ".rst"

    @default("template_name")
    def _template_name_default(self):
        return "rst"

    output_mimetype = "text/restructuredtext"
    export_from_notebook = "reST"

    @property
    def default_config(self):
        c = Config(
            {
                "ExtractOutputPreprocessor": {"enabled": True},
                "HighlightMagicsPreprocessor": {"enabled": True},
            }
        )
        if super().default_config:
            c2 = super().default_config.copy()
            c2.merge(c)
            c = c2
        return c



"""Tests for utils.io"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import io as stdlib_io
import sys
from io import StringIO

from nbconvert.utils.io import unicode_std_stream


def test_UnicodeStdStream():
    # Test wrapping a bytes-level stdout
    stdoutb = stdlib_io.BytesIO()
    stdout = stdlib_io.TextIOWrapper(stdoutb, encoding="ascii")

    orig_stdout = sys.stdout
    sys.stdout = stdout
    try:
        sample = "@łe¶ŧ←"
        unicode_std_stream().write(sample)

        output = stdoutb.getvalue().decode("utf-8")
        assert output == sample
        assert not stdout.closed
    finally:
        sys.stdout = orig_stdout


def test_UnicodeStdStream_nowrap():
    # If we replace stdout with a StringIO, it shouldn't get wrapped.
    orig_stdout = sys.stdout
    sys.stdout = StringIO()
    try:
        assert unicode_std_stream() is sys.stdout
        assert not sys.stdout.closed
    finally:
        sys.stdout = orig_stdout



c = get_config()  # type:ignore  # noqa

# Export all the notebooks in the current directory to the sphinx_howto format.
c.NbConvertApp.notebooks = ["notebook2.ipynb"]
c.NbConvertApp.export_format = "python"



print("import: 'nbconvert'")
import nbconvert

print("import: 'nbconvert.exporters'")
import nbconvert.exporters

print("import: 'nbconvert.filters'")
import nbconvert.filters

print("import: 'nbconvert.postprocessors'")
import nbconvert.postprocessors

print("import: 'nbconvert.preprocessors'")
import nbconvert.preprocessors

print("import: 'nbconvert.resources'")
import nbconvert.resources

print("import: 'nbconvert.utils'")
import nbconvert.utils

print("import: 'nbconvert.writers'")
import nbconvert.writers




"""Tests for the latex preprocessor"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import builtins
from unittest.mock import patch

import pytest

from nbconvert.exporters.exporter import Exporter
from nbconvert.exporters.webpdf import PLAYWRIGHT_INSTALLED, WebPDFExporter

from .base import ExportersTestsBase

real_import = builtins.__import__


class FakeBrowser:
    executable_path: str = ""


def monkey_import_notfound(name, globals_ctx=None, locals_ctx=None, fromlist=(), level=0):
    if name == "playwright.async_api":
        msg = "Fake missing"
        raise ModuleNotFoundError(msg)
    return real_import(name, globals=globals_ctx, locals=locals_ctx, fromlist=fromlist, level=level)


@pytest.mark.skipif(not PLAYWRIGHT_INSTALLED, reason="Playwright not installed")
class TestWebPDFExporter(ExportersTestsBase):
    """Contains test functions for webpdf.py"""

    exporter_class = WebPDFExporter  # type:ignore

    @pytest.mark.network
    def test_export(self):
        """
        Can a TemplateExporter export something?
        """
        (output, resources) = WebPDFExporter(allow_chromium_download=True).from_filename(
            self._get_notebook()
        )
        assert len(output) > 0

    @patch("playwright.async_api._generated.Playwright.chromium", return_value=FakeBrowser())
    def test_webpdf_without_chromium(self, mock_chromium):
        """
        Generate PDFs if chromium not present?
        """
        with pytest.raises(RuntimeError):
            WebPDFExporter(allow_chromium_download=False).from_filename(self._get_notebook())

    @patch("builtins.__import__", monkey_import_notfound)
    def test_webpdf_without_playwright(self):
        """
        Generate PDFs if playwright not installed?
        """
        with pytest.raises(RuntimeError):
            base_exporter = Exporter()
            exporter = WebPDFExporter()
            with open(self._get_notebook(), encoding="utf-8") as f:
                nb = base_exporter.from_file(f, resources={})[0]
                # Have to do this as the very last action as traitlets do dynamic importing often
                exporter.from_notebook_node(nb)



"""
Module with tests for Strings
"""

# -----------------------------------------------------------------------------
# Copyright (c) 2013, the IPython Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import os
import re

from nbconvert.filters.strings import (
    add_anchor,
    add_prompts,
    ascii_only,
    comment_lines,
    get_lines,
    html2text,
    ipython2python,
    posix_path,
    prevent_list_blocks,
    strip_dollars,
    strip_files_prefix,
    wrap_text,
)
from tests.base import TestsBase

# -----------------------------------------------------------------------------
# Class
# -----------------------------------------------------------------------------


class TestStrings(TestsBase):
    def test_wrap_text(self):
        """wrap_text test"""
        test_text = """
        Tush! never tell me; I take it much unkindly
        That thou, Iago, who hast had my purse
        As if the strings were thine, shouldst know of this.
        """
        for length in [30, 5, 1]:
            self._confirm_wrap_text(test_text, length)

    def _confirm_wrap_text(self, text, length):
        for line in wrap_text(text, length).split("\n"):
            assert len(line) <= length

    def test_html2text(self):
        """html2text test"""
        # TODO: More tests
        self.assertEqual(html2text("<name>joe</name>"), "joe")

    def test_add_anchor(self):
        """add_anchor test"""
        # TODO: More tests
        results = add_anchor("<b>Hello World!</b>")
        assert "Hello World!" in results
        assert 'id="' in results
        assert 'class="anchor-link"' in results
        assert "<b" in results
        assert "</b>" in results

    def test_add_anchor_fail(self):
        """add_anchor does nothing when it fails"""
        html = "<h1>Hello <br>World!</h1>"
        results = add_anchor(html)
        self.assertEqual(html, results)

    def test_add_anchor_valid_url_fragment(self):
        """add_anchor creates a valid URL fragment"""
        results = add_anchor(r"<h1>$\pi$ with #s and unicode 中</h1>")
        match = re.search(r'href="#(.*?)"', results)
        assert match
        assert len(match.groups()) == 1
        href = match.groups()[0]

        assert len(href) > 0
        # No invalid characters should be present
        assert "\\" not in href
        assert "#" not in href
        assert "中" not in href

    def test_strip_dollars(self):
        """strip_dollars test"""
        tests = [
            ("", ""),
            ("  ", "  "),
            ("$$", ""),
            ("$H$", "H"),
            ("$He", "He"),
            ("H$el", "H$el"),
            ("Hell$", "Hell"),
            ("Hello", "Hello"),
            ("W$o$rld", "W$o$rld"),
        ]
        for test in tests:
            self._try_strip_dollars(test[0], test[1])

    def _try_strip_dollars(self, test, result):
        self.assertEqual(strip_dollars(test), result)

    def test_strip_files_prefix(self):
        """strip_files_prefix test"""
        tests = [
            ("", ""),
            ("/files", "/files"),
            ('test="/files"', 'test="/files"'),
            ("My files are in `files/`", "My files are in `files/`"),
            (
                '<a href="files/test.html">files/test.html</a>',
                '<a href="test.html">files/test.html</a>',
            ),
            (
                '<a href="/files/test.html">files/test.html</a>',
                '<a href="test.html">files/test.html</a>',
            ),
            (
                "<a href='files/test.html'>files/test.html</a>",
                "<a href='test.html'>files/test.html</a>",
            ),
            ('<img src="files/url/location.gif">', '<img src="url/location.gif">'),
            ('<img src="/files/url/location.gif">', '<img src="url/location.gif">'),
            ("hello![caption]", "hello![caption]"),
            ("hello![caption](/url/location.gif)", "hello![caption](/url/location.gif)"),
            ("hello![caption](url/location.gif)", "hello![caption](url/location.gif)"),
            ("hello![caption](url/location.gif)", "hello![caption](url/location.gif)"),
            ("hello![caption](files/url/location.gif)", "hello![caption](url/location.gif)"),
            ("hello![caption](/files/url/location.gif)", "hello![caption](url/location.gif)"),
            ("hello [text](/files/url/location.gif)", "hello [text](url/location.gif)"),
            ("hello [text space](files/url/location.gif)", "hello [text space](url/location.gif)"),
        ]
        for test in tests:
            self._try_files_prefix(test[0], test[1])

    def _try_files_prefix(self, test, result):
        self.assertEqual(strip_files_prefix(test), result)

    def test_comment_lines(self):
        """comment_lines test"""
        for line in comment_lines("hello\nworld\n!").split("\n"):
            assert line.startswith("# ")
        for line in comment_lines("hello\nworld\n!", "beep").split("\n"):
            assert line.startswith("beep")

    def test_get_lines(self):
        """get_lines test"""
        text = "hello\nworld\n!"
        self.assertEqual(get_lines(text, start=1), "world\n!")
        self.assertEqual(get_lines(text, end=2), "hello\nworld")
        self.assertEqual(get_lines(text, start=2, end=5), "!")
        self.assertEqual(get_lines(text, start=-2), "world\n!")

    def test_ipython2python(self):
        """ipython2python test"""
        # TODO: More tests
        results = ipython2python('%%pylab\nprint("Hello-World")').replace("u'", "'")
        self.fuzzy_compare(
            results.replace(r"\n", ""),
            "get_ipython().run_cell_magic('pylab', '', 'print(\"Hello-World\")')",
            ignore_spaces=True,
            ignore_newlines=True,
        )

    def test_posix_path(self):
        """posix_path test"""
        path_list = ["foo", "bar"]
        expected = "/".join(path_list)
        native = os.path.join(*path_list)
        filtered = posix_path(native)
        self.assertEqual(filtered, expected)

    def test_add_prompts(self):
        """add_prompts test"""
        text1 = """for i in range(10):\n  i += 1\n  print i"""
        text2 = """>>> for i in range(10):\n...   i += 1\n...   print i"""
        self.assertEqual(text2, add_prompts(text1))

    def test_prevent_list_blocks(self):
        """prevent_list_blocks test"""
        tests = [
            ("1. arabic point", "1\\. arabic point"),
            ("* bullet asterisk", "\\* bullet asterisk"),
            ("+ bullet Plus Sign", "\\+ bullet Plus Sign"),
            ("- bullet Hyphen-Minus", "\\- bullet Hyphen-Minus"),
            ("  1. spaces + arabic point", "  1\\. spaces + arabic point"),
        ]
        for test in tests:
            self.assertEqual(prevent_list_blocks(test[0]), test[1])

    def test_ascii_only(self):
        """ascii only test"""
        tests = [
            ("", ""),
            ("  ", "  "),
            ("Hello", "Hello"),
            ("Hello 中文", "Hello ??"),
        ]
        for test in tests:
            self.assertEqual(test[1], ascii_only(test[0]))



"""Module with tests for DataTypeFilter"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.
import pytest

from nbconvert.filters.datatypefilter import DataTypeFilter
from tests.base import TestsBase


class TestDataTypeFilter(TestsBase):
    """Contains test functions for datatypefilter.py"""

    def test_constructor(self):
        """Can an instance of a DataTypeFilter be created?"""
        DataTypeFilter()

    def test_junk_types(self):
        """Can the DataTypeFilter pickout a useful type from a dict with junk types as keys?"""
        filter_ = DataTypeFilter()
        assert "image/png" in filter_({"hair": "1", "water": 2, "image/png": 3, "rock": 4.0})
        assert "application/pdf" in filter_(
            {
                "application/pdf": "file_path",
                "hair": 2,
                "water": "yay",
                "png": "not a png",
                "rock": "is a rock",
            }
        )

        with pytest.warns(UserWarning):
            self.assertEqual(
                filter_(
                    {"hair": "this is not", "water": "going to return anything", "rock": "or is it"}
                ),
                [],
            )

    def test_null(self):
        """Will the DataTypeFilter fail if no types are passed in?"""
        filter_ = DataTypeFilter()
        with pytest.warns(UserWarning):
            self.assertEqual(filter_({}), [])



"""This preprocessor detect cells using a different language through
magic extensions such as `%%R` or `%%octave`. Cell's metadata is marked
so that the appropriate highlighter can be used in the `highlight`
filter.
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.


import re

from traitlets import Dict

from .base import Preprocessor


class HighlightMagicsPreprocessor(Preprocessor):
    """
    Detects and tags code cells that use a different languages than Python.
    """

    # list of magic language extensions and their associated pygment lexers
    default_languages = Dict(
        {
            "%%R": "r",
            "%%bash": "bash",
            "%%cython": "cython",
            "%%javascript": "javascript",
            "%%julia": "julia",
            "%%latex": "latex",
            "%%octave": "octave",
            "%%perl": "perl",
            "%%ruby": "ruby",
            "%%sh": "sh",
            "%%sql": "sql",
        }
    )

    # user defined language extensions
    languages = Dict(
        help=(
            "Syntax highlighting for magic's extension languages. "
            "Each item associates a language magic extension such as %%R, "
            "with a pygments lexer such as r."
        )
    ).tag(config=True)

    def __init__(self, config=None, **kw):
        """Public constructor"""

        super().__init__(config=config, **kw)

        # Update the default languages dict with the user configured ones
        self.default_languages.update(self.languages)

        # build a regular expression to catch language extensions and choose
        # an adequate pygments lexer
        any_language = "|".join(self.default_languages.keys())
        self.re_magic_language = re.compile(rf"^\s*({any_language})\s+")

    def which_magic_language(self, source):
        """
        When a cell uses another language through a magic extension,
        the other language is returned.
        If no language magic is detected, this function returns None.

        Parameters
        ----------
        source: str
            Source code of the cell to highlight
        """

        m = self.re_magic_language.match(source)

        if m:
            # By construction of the re, the matched language must be in the
            # languages dictionary
            return self.default_languages[m.group(1)]
        else:
            return None

    def preprocess_cell(self, cell, resources, cell_index):
        """
        Tags cells using a magic extension language

        Parameters
        ----------
        cell : NotebookNode cell
            Notebook cell being processed
        resources : dictionary
            Additional resources used in the conversion process.  Allows
            preprocessors to pass variables into the Jinja engine.
        cell_index : int
            Index of the cell being processed (see base.py)
        """

        # Only tag code cells
        if cell.cell_type == "code":
            magic_language = self.which_magic_language(cell.source)
            if magic_language:
                cell["metadata"]["magics_language"] = magic_language
        return cell, resources



"""Export to PDF via a headless browser"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from traitlets import Bool

from .qt_exporter import QtExporter


class QtPDFExporter(QtExporter):
    """Writer designed to write to PDF files.

    This inherits from :class:`HTMLExporter`. It creates the HTML using the
    template machinery, and then uses pyqtwebengine to create a pdf.
    """

    export_from_notebook = "PDF via HTML"
    format = "pdf"  # noqa

    paginate = Bool(  # type:ignore[assignment]
        True,
        help="""
        Split generated notebook into multiple pages.

        If False, a PDF with one long page will be generated.

        Set to True to match behavior of LaTeX based PDF generator
        """,
    ).tag(config=True)



"""Test Pandoc module"""
# -----------------------------------------------------------------------------
#  Copyright (C) 2014 The IPython Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import os
import warnings

from nbconvert.utils import pandoc
from tests.base import TestsBase
from tests.testutils import onlyif_cmds_exist

# -----------------------------------------------------------------------------
# Classes and functions
# -----------------------------------------------------------------------------


class TestPandoc(TestsBase):
    """Collection of Pandoc tests"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.original_env = os.environ.copy()

    def setUp(self):
        super().setUp()
        pandoc.check_pandoc_version._cached = None  # type:ignore

    @onlyif_cmds_exist("pandoc")
    def test_pandoc_available(self):
        """Test behaviour that pandoc functions raise PandocMissing as documented"""
        pandoc.clean_cache()

        os.environ["PATH"] = ""
        with self.assertRaises(pandoc.PandocMissing):
            pandoc.get_pandoc_version()
        with self.assertRaises(pandoc.PandocMissing):
            pandoc.check_pandoc_version()
        with self.assertRaises(pandoc.PandocMissing):
            pandoc.pandoc("", "markdown", "html")

        # original_env["PATH"] should contain pandoc
        os.environ["PATH"] = self.original_env["PATH"]
        with warnings.catch_warnings(record=True) as w:
            pandoc.get_pandoc_version()
            pandoc.check_pandoc_version()
            pandoc.pandoc("", "markdown", "html")
        self.assertEqual(w, [])

    @onlyif_cmds_exist("pandoc")
    def test_minimal_version(self):
        pandoc._minimal_version = "120.0"
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # call it twice to verify the cached value is used
            assert not pandoc.check_pandoc_version()
            assert not pandoc.check_pandoc_version()
        # only one warning after two calls, due to cache
        self.assertEqual(len(w), 1)
        # clear cache
        pandoc.check_pandoc_version._cached = None  # type:ignore
        pandoc._minimal_version = pandoc.get_pandoc_version()
        assert pandoc.check_pandoc_version()


def pandoc_function_raised_missing(f, *args, **kwargs):
    try:
        f(*args, **kwargs)
    except pandoc.PandocMissing:
        return True
    else:
        return False



from nbconvert.writers.base import WriterBase


class HelloWriter(WriterBase):
    def write(self, output, resources, notebook_name=None, **kw):
        with open("hello.txt", "w") as outfile:
            outfile.write("hello world")



"""
Module with tests for stdout
"""

# -----------------------------------------------------------------------------
# Copyright (c) 2013, the IPython Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import sys
from io import StringIO

from nbconvert.writers.stdout import StdoutWriter
from tests.base import TestsBase

# -----------------------------------------------------------------------------
# Class
# -----------------------------------------------------------------------------


class TestStdout(TestsBase):
    """Contains test functions for stdout.py"""

    def test_output(self):
        """Test stdout writer output."""

        # Capture the stdout.  Remember original.
        stdout = sys.stdout
        stream = StringIO()
        sys.stdout = stream

        # Create stdout writer, test output
        writer = StdoutWriter()
        writer.write("ax", {"b": "c"})
        output = stream.getvalue()
        self.fuzzy_compare(output, "ax")

        # Revert stdout
        sys.stdout = stdout



"""
Convert between any two formats using pandoc,
and related filters
"""
import os

from pandocfilters import Image, applyJSONFilters  # type:ignore[import-untyped]

from nbconvert.utils.base import NbConvertBase
from nbconvert.utils.pandoc import pandoc

__all__ = ["ConvertExplicitlyRelativePaths", "convert_pandoc"]


def convert_pandoc(source, from_format, to_format, extra_args=None):
    """Convert between any two formats using pandoc.

    This function will raise an error if pandoc is not installed.
    Any error messages generated by pandoc are printed to stderr.

    Parameters
    ----------
    source : string
        Input string, assumed to be valid in from_format.
    from_format : string
        Pandoc format of source.
    to_format : string
        Pandoc format for output.

    Returns
    -------
    out : string
        Output as returned by pandoc.
    """
    return pandoc(source, from_format, to_format, extra_args=extra_args)


# When converting to pdf, explicitly relative references
# like "./" and "../" doesn't work with TEXINPUTS.
# So we need to convert them to absolute paths.
# See https://github.com/jupyter/nbconvert/issues/1998
class ConvertExplicitlyRelativePaths(NbConvertBase):
    """A converter that handles relative path references."""

    def __init__(self, texinputs=None, **kwargs):
        """Initialize the converter."""
        # texinputs should be the directory of the notebook file
        self.nb_dir = os.path.abspath(texinputs) if texinputs else ""
        self.ancestor_dirs = self.nb_dir.split("/")
        super().__init__(**kwargs)

    def __call__(self, source):
        """Invoke the converter."""
        # If this is not set for some reason, we can't do anything,
        if self.nb_dir:
            return applyJSONFilters([self.action], source)
        return source

    def action(self, key, value, frmt, meta):
        """Perform the action."""
        # Convert explicitly relative paths:
        # ./path -> path  (This should be visible to the latex engine since TEXINPUTS already has .)
        # ../path -> /abs_path
        # assuming all relative references are at the start of a given path
        if key == "Image":
            # Image seems to have this composition, according to https://github.com/jgm/pandoc-types
            attr, caption, [filename, typedef] = value

            if filename[:2] == "./":
                filename = filename[2:]
            elif filename[:3] == "../":
                n_up = 0
                while filename[:3] == "../":
                    n_up += 1
                    filename = filename[3:]
                ancestors = "/".join(self.ancestor_dirs[:-n_up]) + "/"
                filename = ancestors + filename
            return Image(attr, caption, [filename, typedef])
        # If not image, return "no change"
        return None



"""
Module with tests for debug
"""

# -----------------------------------------------------------------------------
# Copyright (c) 2013, the IPython Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import sys
from io import StringIO

from nbconvert.writers.debug import DebugWriter
from tests.base import TestsBase

# -----------------------------------------------------------------------------
# Class
# -----------------------------------------------------------------------------


class TestDebug(TestsBase):
    """Contains test functions for debug.py"""

    def test_output(self):
        """Test debug writer output."""

        # Capture the stdout.  Remember original.
        stdout = sys.stdout
        stream = StringIO()
        sys.stdout = stream

        # Create stdout writer, get output
        writer = DebugWriter()
        writer.write("aaa", {"outputs": {"bbb": "ccc"}})
        output = stream.getvalue()

        # Check output.  Make sure resources dictionary is dumped, but nothing
        # else.
        assert "aaa" not in output
        assert "bbb" in output
        assert "ccc" in output

        # Revert stdout
        sys.stdout = stdout



"""Tests for the HighlightMagics preprocessor"""

from nbconvert.preprocessors.highlightmagics import HighlightMagicsPreprocessor

from .base import PreprocessorTestsBase


class TestHighlightMagics(PreprocessorTestsBase):
    """Contains test functions for highlightmagics.py"""

    def build_preprocessor(self):
        """Make an instance of a preprocessor"""
        preprocessor = HighlightMagicsPreprocessor()
        preprocessor.enabled = True
        return preprocessor

    def test_constructor(self):
        """Can a HighlightMagicsPreprocessor be constructed?"""
        self.build_preprocessor()

    def test_tagging(self):
        """Test the HighlightMagicsPreprocessor tagging"""
        nb = self.build_notebook()
        res = self.build_resources()
        preprocessor = self.build_preprocessor()
        nb.cells[0].source = """%%R -i x,y -o XYcoef
            lm.fit <- lm(y~x)
            par(mfrow=c(2,2))
            print(summary(lm.fit))
            plot(lm.fit)
            XYcoef <- coef(lm.fit)"""

        nb, res = preprocessor(nb, res)

        assert "magics_language" in nb.cells[0]["metadata"]

        self.assertEqual(nb.cells[0]["metadata"]["magics_language"], "r")

    def test_no_false_positive(self):
        """Test that HighlightMagicsPreprocessor does not tag false positives"""
        nb = self.build_notebook()
        res = self.build_resources()
        preprocessor = self.build_preprocessor()
        nb.cells[0].source = """# this should not be detected
                print(\"""
                %%R -i x, y
                \""")"""
        nb, res = preprocessor(nb, res)

        assert "magics_language" not in nb.cells[0]["metadata"]



"""
Contains debug writer.
"""

from pprint import pprint

from .base import WriterBase

# -----------------------------------------------------------------------------
# Copyright (c) 2013, the IPython Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------


class DebugWriter(WriterBase):
    """Consumes output from nbconvert export...() methods and writes useful
    debugging information to the stdout.  The information includes a list of
    resources that were extracted from the notebook(s) during export."""

    def write(self, output, resources, notebook_name="notebook", **kw):
        """
        Consume and write Jinja output.

        See base for more...
        """

        if isinstance(resources["outputs"], dict):
            print("outputs extracted from %s" % notebook_name)
            print("-" * 80)
            pprint(resources["outputs"], indent=2, width=70)  # noqa
        else:
            print("no outputs extracted from %s" % notebook_name)
        print("=" * 80)



"""Test NbConvertApp"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import os
from tempfile import TemporaryDirectory

import nbformat
import pytest
from flaky import flaky  # type:ignore
from traitlets.tests.utils import check_help_all_output

from nbconvert.exporters import HTMLExporter
from nbconvert.exporters.webpdf import PLAYWRIGHT_INSTALLED
from nbconvert.postprocessors import PostProcessorBase

from .base import TestsBase
from .testutils import onlyif_cmds_exist

# -----------------------------------------------------------------------------
# Classes and functions
# -----------------------------------------------------------------------------


class DummyPost(PostProcessorBase):
    def postprocess(self, filename):
        print("Dummy:%s" % filename)


class TestNbConvertApp(TestsBase):
    """Collection of NbConvertApp tests"""

    def test_notebook_help(self):
        """Will help show if no notebooks are specified?"""
        with self.create_temp_cwd():
            out, err = self.nbconvert("--log-level 0", ignore_return_code=True)
            self.assertIn("--help-all", out)

    def test_help_output(self):
        """ipython nbconvert --help-all works"""
        check_help_all_output("nbconvert")

    def test_glob(self):
        """
        Do search patterns work for notebook names?
        """
        with self.create_temp_cwd(["notebook*.ipynb"]):
            self.nbconvert("--to python *.ipynb --log-level 0")
            assert os.path.isfile("notebook1.py")
            assert os.path.isfile("notebook2.py")

    def test_glob_subdir(self):
        """
        Do search patterns work for subdirectory notebook names?
        """
        with self.create_temp_cwd():
            self.copy_files_to(["notebook*.ipynb"], "subdir/")
            self.nbconvert("--to python --log-level 0 " + os.path.join("subdir", "*.ipynb"))
            assert os.path.isfile(os.path.join("subdir", "notebook1.py"))
            assert os.path.isfile(os.path.join("subdir", "notebook2.py"))

    def test_build_dir(self):
        """build_directory affects export location"""
        with self.create_temp_cwd():
            self.copy_files_to(["notebook*.ipynb"], "subdir/")
            self.nbconvert(
                "--to python --log-level 0 --output-dir . " + os.path.join("subdir", "*.ipynb")
            )
            assert os.path.isfile("notebook1.py")
            assert os.path.isfile("notebook2.py")

    def test_convert_full_qualified_name(self):
        """
        Test that nbconvert can convert file using a full qualified name for a
        package, import and use it.
        """
        with self.create_temp_cwd():
            self.copy_files_to(["notebook*.ipynb"], "subdir")
            self.nbconvert(
                "--to tests.fake_exporters.MyExporter --log-level 0 "
                + os.path.join("subdir", "*.ipynb")
            )
            assert os.path.isfile(os.path.join("subdir", "notebook1.test_ext"))
            assert os.path.isfile(os.path.join("subdir", "notebook2.test_ext"))

    def test_explicit(self):
        """
        Do explicit notebook names work?
        """
        with self.create_temp_cwd(["notebook*.ipynb"]):
            self.nbconvert("--log-level 0 --to python notebook2")
            assert not os.path.isfile("notebook1.py")
            assert os.path.isfile("notebook2.py")

    def test_clear_output(self):
        """
        Can we clear outputs?
        """
        with self.create_temp_cwd(["notebook*.ipynb"]):
            self.nbconvert("--clear-output notebook1")
            assert os.path.isfile("notebook1.ipynb")
            with open("notebook1.ipynb", encoding="utf8") as f:
                nb = nbformat.read(f, 4)
                for cell in nb.cells:
                    # Skip markdown cells
                    if "outputs" in cell:
                        assert cell.outputs == []

    def test_absolute_template_file(self):
        """--template-file '/path/to/template.tpl'"""
        with self.create_temp_cwd(["notebook*.ipynb"]), TemporaryDirectory() as td:
            template = os.path.join(td, "mytemplate.tpl")
            test_output = "success!"
            with open(template, "w") as f:
                f.write(test_output)
            self.nbconvert("--log-level 0 notebook2 --to html --template-file %s" % template)
            assert os.path.isfile("notebook2.html")
            with open("notebook2.html", encoding="utf8") as f:
                text = f.read()
            assert text == test_output

    def test_relative_template_file(self):
        """Test --template-file 'relative/path.tpl'"""
        with self.create_temp_cwd(["notebook*.ipynb"]):
            os.mkdir("relative")
            template = os.path.join("relative", "path.tpl")
            test_output = "success!"
            with open(template, "w") as f:
                f.write(test_output)
            self.nbconvert("--log-level 0 notebook2 --to html --template-file %s" % template)
            assert os.path.isfile("notebook2.html")
            with open("notebook2.html", encoding="utf8") as f:
                text = f.read()
            assert text == test_output

    @onlyif_cmds_exist("pandoc", "xelatex")
    def test_filename_spaces(self):
        """
        Generate PDFs with graphics if notebooks have spaces in the name?
        """
        with self.create_temp_cwd(["notebook2.ipynb"]):
            os.rename("notebook2.ipynb", "notebook with spaces.ipynb")
            self.nbconvert(
                "--log-level 0 --to pdf"
                ' "notebook with spaces"'
                " --PDFExporter.latex_count=1"
                " --PDFExporter.verbose=True"
            )
            assert os.path.isfile("notebook with spaces.pdf")

    @flaky
    @pytest.mark.network
    @pytest.mark.skipif(not PLAYWRIGHT_INSTALLED, reason="Playwright not installed")
    def test_webpdf_with_chromium(self):
        """
        Generate PDFs if chromium allowed to be downloaded?
        """
        with self.create_temp_cwd(["notebook2.ipynb"]):
            self.nbconvert('--to webpdf --allow-chromium-download "notebook2"')
            assert os.path.isfile("notebook2.pdf")

    @onlyif_cmds_exist("pandoc", "xelatex")
    def test_pdf(self):
        """
        Check to see if pdfs compile, even if strikethroughs are included.
        """
        with self.create_temp_cwd(["notebook2.ipynb"]):
            self.nbconvert(
                "--log-level 0 --to pdf"
                ' "notebook2"'
                " --PDFExporter.latex_count=1"
                " --PDFExporter.verbose=True"
            )
            assert os.path.isfile("notebook2.pdf")

    def test_post_processor(self):
        """Do post processors work?"""
        with self.create_temp_cwd(["notebook1.ipynb"]):
            out, err = self.nbconvert(
                "--log-level 0 --to python notebook1 --post tests.test_nbconvertapp.DummyPost"
            )
            self.assertIn("Dummy:notebook1.py", out)

    @onlyif_cmds_exist("pandoc")
    def test_spurious_cr(self):
        """Check for extra CR characters"""
        with self.create_temp_cwd(["notebook2.ipynb"]):
            self.nbconvert("--log-level 0 --to latex notebook2")
            assert os.path.isfile("notebook2.tex")
            with open("notebook2.tex") as f:
                tex = f.read()
            self.nbconvert("--log-level 0 --to html notebook2")
            assert os.path.isfile("notebook2.html")
            with open("notebook2.html", encoding="utf8") as f:
                html = f.read()
        self.assertEqual(tex.count("\r"), tex.count("\r\n"))
        self.assertEqual(html.count("\r"), html.count("\r\n"))

    @onlyif_cmds_exist("pandoc")
    def test_png_base64_html_ok(self):
        """Is embedded png data well formed in HTML?"""
        with self.create_temp_cwd(["notebook2.ipynb"]):
            self.nbconvert("--log-level 0 --to HTML notebook2.ipynb --template lab")
            assert os.path.isfile("notebook2.html")
            with open("notebook2.html", encoding="utf8") as f:
                assert "data:image/png;base64,b'" not in f.read()

    @onlyif_cmds_exist("pandoc")
    def test_template(self):
        """
        Do export templates work?
        """
        with self.create_temp_cwd(["notebook2.ipynb"]):
            self.nbconvert("--log-level 0 --to slides notebook2.ipynb")
            assert os.path.isfile("notebook2.slides.html")
            with open("notebook2.slides.html", encoding="utf8") as f:
                assert "/reveal.css" in f.read()

    def test_output_ext(self):
        """test --output=outputfile[.ext]"""
        with self.create_temp_cwd(["notebook1.ipynb"]):
            self.nbconvert("--log-level 0 --to python notebook1.ipynb --output nb.py")
            assert os.path.exists("nb.py")

            self.nbconvert("--log-level 0 --to python notebook1.ipynb --output nb2")
            assert os.path.exists("nb2.py")

    def test_glob_explicit(self):
        """
        Can a search pattern be used along with matching explicit notebook names?
        """
        with self.create_temp_cwd(["notebook*.ipynb"]):
            self.nbconvert("--log-level 0 --to python *.ipynb notebook1.ipynb notebook2.ipynb")
            assert os.path.isfile("notebook1.py")
            assert os.path.isfile("notebook2.py")

    def test_explicit_glob(self):
        """
        Can explicit notebook names be used and then a matching search pattern?
        """
        with self.create_temp_cwd(["notebook*.ipynb"]):
            self.nbconvert("--log-level 0 --to=python notebook1.ipynb notebook2.ipynb *.ipynb")
            assert os.path.isfile("notebook1.py")
            assert os.path.isfile("notebook2.py")

    def test_default_config(self):
        """
        Does the default config work?
        """
        with self.create_temp_cwd(["notebook*.ipynb", "jupyter_nbconvert_config.py"]):
            self.nbconvert("--log-level 0 --config jupyter_nbconvert_config.py")
            assert os.path.isfile("notebook1.py")
            assert not os.path.isfile("notebook2.py")

    def test_override_config(self):
        """
        Can the default config be overridden?
        """
        with self.create_temp_cwd(
            ["notebook*.ipynb", "jupyter_nbconvert_config.py", "override.py"]
        ):
            self.nbconvert('--log-level 0 --config="override.py"')
            assert not os.path.isfile("notebook1.py")
            assert os.path.isfile("notebook2.py")

    def test_accents_in_filename(self):
        """
        Can notebook names include accents?
        """
        with self.create_temp_cwd():
            self.create_empty_notebook("nb1_análisis.ipynb")
            self.nbconvert("--log-level 0 --to Python nb1_*")
            assert os.path.isfile("nb1_análisis.py")

    @onlyif_cmds_exist("xelatex", "pandoc")
    def test_filename_accent_pdf(self):
        """
        Generate PDFs if notebooks have an accent in their name?
        """
        with self.create_temp_cwd():
            self.create_empty_notebook("nb1_análisis.ipynb")
            self.nbconvert(
                '--log-level 0 --to pdf "nb1_*"'
                " --PDFExporter.latex_count=1"
                " --PDFExporter.verbose=True"
            )
            assert os.path.isfile("nb1_análisis.pdf")

    @pytest.mark.skipif(os.name == "nt", reason="CLI parsing does not work the same on Windows")
    def test_cwd_plugin(self):
        """
        Verify that an extension in the cwd can be imported.
        """
        with self.create_temp_cwd(["hello.py"]):
            self.create_empty_notebook("empty.ipynb")
            assert os.path.isfile("hello.py")
            self.nbconvert("empty --to html --NbConvertApp.writer_class='hello.HelloWriter'")
            assert os.path.isfile("hello.txt")

    def test_output_suffix(self):
        """
        Verify that the output suffix is applied
        """
        with self.create_temp_cwd():
            self.create_empty_notebook("empty.ipynb")
            self.nbconvert("empty.ipynb --to notebook")
            assert os.path.isfile("empty.nbconvert.ipynb")

    def test_different_build_dir(self):
        """
        Verify that the output suffix is not applied
        """
        with self.create_temp_cwd():
            self.create_empty_notebook("empty.ipynb")
            os.mkdir("output")
            self.nbconvert("empty.ipynb --to notebook --FilesWriter.build_directory=output")
            assert os.path.isfile("output/empty.ipynb")

    def test_inplace(self):
        """
        Verify that the notebook is converted in place
        """
        with self.create_temp_cwd():
            self.create_empty_notebook("empty.ipynb")
            self.nbconvert("empty.ipynb --inplace")
            assert os.path.isfile("empty.ipynb")
            assert not os.path.isfile("empty.nbconvert.ipynb")
            assert not os.path.isfile("empty.html")

    def test_no_prompt(self):
        """
        Verify that the html has no prompts when given --no-prompt.
        """
        with self.create_temp_cwd(["notebook1.ipynb"]):
            self.nbconvert("notebook1.ipynb --log-level 0 --no-prompt --to html")
            assert os.path.isfile("notebook1.html")
            with open("notebook1.html", encoding="utf8") as f:
                text = f.read()
                assert "In\xa0[" not in text
                assert "Out[6]" not in text
            self.nbconvert("notebook1.ipynb --log-level 0 --to html")
            assert os.path.isfile("notebook1.html")
            with open("notebook1.html", encoding="utf8") as f:
                text2 = f.read()
                assert "In\xa0[" in text2
                assert "Out[6]" in text2

    def test_cell_tag_output(self):
        """
        Verify that the html has tags in cell attributes if they exist.
        """
        with self.create_temp_cwd(["notebook_tags.ipynb"]):
            self.nbconvert("notebook_tags.ipynb --log-level 0 --to html")
            assert os.path.isfile("notebook_tags.html")
            with open("notebook_tags.html", encoding="utf8") as f:
                text = f.read()
                assert "celltag_mycelltag celltag_mysecondcelltag" in text
                assert "celltag_mymarkdowncelltag" in text

    def test_no_input(self):
        """
        Verify that the html has no input when given --no-input.
        """
        input_content_html = (
            '<span class="n">x</span>'
            '<span class="p">,</span>'
            '<span class="n">y</span>'
            '<span class="p">,</span>'
            '<span class="n">z</span> '
            '<span class="o">=</span> '
            '<span class="n">symbols</span>'
            '<span class="p">(</span>'
            "<span class=\"s1\">'x y z'</span>"
            '<span class="p">)</span>'
        )
        for no_input_flag in (False, True):
            with self.create_temp_cwd(["notebook1.ipynb"]):
                command = "notebook1.ipynb --log-level 0 --to html"
                if no_input_flag:
                    command += " --no-input"
                self.nbconvert(command)
                assert os.path.isfile("notebook1.html")

                with open("notebook1.html", encoding="utf8") as f:
                    text = f.read()
                    assert no_input_flag == ("In\xa0[" not in text)
                    assert no_input_flag == ("Out[6]" not in text)
                    assert no_input_flag == (input_content_html not in text)

    def test_allow_errors(self):
        """
        Verify that conversion is aborted with '--execute' if an error is
        encountered, but that conversion continues if '--allow-errors' is
        used in addition.
        """
        with self.create_temp_cwd(["notebook3*.ipynb"]):
            # Convert notebook containing a cell that raises an error,
            # both without and with cell execution enabled.
            output1, _ = self.nbconvert(
                "--to markdown --stdout notebook3*.ipynb"
            )  # no cell execution
            output2, _ = self.nbconvert(
                "--to markdown --allow-errors --stdout notebook3*.ipynb"
            )  # no cell execution; --allow-errors should have no effect
            output3, _ = self.nbconvert(
                "--execute --allow-errors --to markdown --stdout notebook3*.ipynb"
            )  # with cell execution; errors are allowed

            # Un-executed outputs should not contain either
            # of the two numbers computed in the notebook.
            assert "23" not in output1
            assert "42" not in output1
            assert "23" not in output2
            assert "42" not in output2

            # Executed output should contain both numbers.
            assert "23" in output3
            assert "42" in output3

            # Executing the notebook should raise an exception if --allow-errors is not specified
            with pytest.raises(OSError):
                self.nbconvert("--execute --to markdown --stdout notebook3*.ipynb")

    def test_errors_print_traceback(self):
        """
        Verify that the stderr output contains the traceback of the cell execution exception.
        """
        with self.create_temp_cwd(["notebook3_with_errors.ipynb"]):
            _, error_output = self.nbconvert(
                "--execute --to markdown --stdout notebook3_with_errors.ipynb",
                ignore_return_code=True,
            )
            assert 'print("Some text before the error")' in error_output
            assert 'raise RuntimeError("This is a deliberate exception")' in error_output
            assert "RuntimeError" in error_output
            assert "This is a deliberate exception" in error_output

    def test_fenced_code_blocks_markdown(self):
        """
        Verify that input cells use fenced code blocks with the language
        name in nb.metadata.kernelspec.language, if that exists
        """
        with self.create_temp_cwd(["notebook1*.ipynb"]):
            # this notebook doesn't have nb.metadata.kernelspec, so it should
            # just do a fenced code block, with no language
            output1, _ = self.nbconvert("--to markdown --stdout notebook1.ipynb")
            assert "```python" not in output1  # shouldn't have language
            assert "```" in output1  # but should have fenced blocks

        with self.create_temp_cwd(["notebook_jl*.ipynb"]):
            output2, _ = self.nbconvert("--to markdown --stdout notebook_jl.ipynb")
            assert "```julia" in output2  # shouldn't have language
            assert "```" in output2  # but should also plain ``` to close cell

    def test_convert_from_stdin_to_stdout(self):
        """
        Verify that conversion can be done via stdin to stdout
        """
        with self.create_temp_cwd(["notebook1.ipynb"]):
            with open("notebook1.ipynb") as f:
                notebook = f.read().encode()
                output1, _ = self.nbconvert("--to markdown --stdin --stdout", stdin=notebook)
            assert "```python" not in output1  # shouldn't have language
            assert "```" in output1  # but should have fenced blocks

    def test_convert_from_stdin(self):
        """
        Verify that conversion can be done via stdin.
        """
        with self.create_temp_cwd(["notebook1.ipynb"]):
            with open("notebook1.ipynb") as f:
                notebook = f.read().encode()
                self.nbconvert("--to markdown --stdin", stdin=notebook)
            assert os.path.isfile("notebook.md")  # default name for stdin input
            with open("notebook.md", encoding="utf8") as f:
                output1 = f.read()
                assert "```python" not in output1  # shouldn't have language
                assert "```" in output1  # but should have fenced blocks

    @onlyif_cmds_exist("pandoc", "xelatex")
    def test_linked_images(self):
        """
        Generate PDFs with an image linked in a markdown cell
        """
        with self.create_temp_cwd(["latex-linked-image.ipynb", "testimage.png"]):
            self.nbconvert("--to pdf latex-linked-image.ipynb")
            assert os.path.isfile("latex-linked-image.pdf")

    @onlyif_cmds_exist("pandoc")
    def test_embedded_jpeg(self):
        """
        Verify that latex conversion succeeds
        with a notebook with an embedded .jpeg
        """
        with self.create_temp_cwd(["notebook4_jpeg.ipynb", "containerized_deployments.jpeg"]):
            self.nbconvert("--to latex notebook4_jpeg.ipynb")
            assert os.path.isfile("notebook4_jpeg.tex")

    @onlyif_cmds_exist("pandoc")
    def test_markdown_display_priority(self):
        """
        Check to see if markdown conversion embeds PNGs,
        even if an (unsupported) PDF is present.
        """
        with self.create_temp_cwd(["markdown_display_priority.ipynb"]):
            self.nbconvert('--log-level 0 --to markdown "markdown_display_priority.ipynb"')
            assert os.path.isfile("markdown_display_priority.md")
            with open("markdown_display_priority.md", encoding="utf8") as f:
                markdown_output = f.read()
                assert (
                    "markdown_display_priority_files/markdown_display_priority_0_1.png"
                ) in markdown_output

    @onlyif_cmds_exist("pandoc")
    def test_write_figures_to_custom_path(self):
        """
        Check if figure files are copied to configured path.
        """

        def fig_exists(path):
            return len(os.listdir(path)) > 0

        # check absolute path
        with self.create_temp_cwd(["notebook4_jpeg.ipynb", "containerized_deployments.jpeg"]):
            output_dir = TemporaryDirectory()
            path = os.path.join(output_dir.name, "files")
            self.nbconvert(
                "--log-level 0 notebook4_jpeg.ipynb --to rst "
                f"--NbConvertApp.output_files_dir={path}"
            )
            assert fig_exists(path)
            output_dir.cleanup()

        # check relative path
        with self.create_temp_cwd(["notebook4_jpeg.ipynb", "containerized_deployments.jpeg"]):
            self.nbconvert(
                "--log-level 0 notebook4_jpeg.ipynb --to rst "
                "--NbConvertApp.output_files_dir=output"
            )
            assert fig_exists("output")

        # check default path with notebook name
        with self.create_temp_cwd(["notebook4_jpeg.ipynb", "containerized_deployments.jpeg"]):
            self.nbconvert("--log-level 0 notebook4_jpeg.ipynb --to rst")
            assert fig_exists("notebook4_jpeg_files")

    def test_widgets_from_nbconvert(self):
        """Check jupyter widgets URL"""

        with self.create_temp_cwd(["Widget_List.ipynb"]):
            self.nbconvert("Widget_List.ipynb --log-level 0 --to html")
            assert os.path.isfile("Widget_List.html")
            with open("Widget_List.html", encoding="utf8") as f:
                text = f.read()
                assert (
                    "var widgetRendererSrc = 'https://unpkg.com/@jupyter-widgets/html-manager@*/dist/embed-amd.js';"
                    in text
                )

    def test_widgets_from_htmlexporter(self):
        """Check jupyter widgets URL"""
        with self.create_temp_cwd(["Widget_List.ipynb"]) as tmpdir:
            with open(os.path.join(tmpdir, "Widget_List.ipynb")) as f:
                nb = nbformat.read(f, 4)

            output, _ = HTMLExporter().from_notebook_node(nb)

            assert (
                "var widgetRendererSrc = 'https://unpkg.com/@jupyter-widgets/html-manager@*/dist/embed-amd.js';"
                in output
            )

    def test_not_embedding_images_htmlexporter(self):
        """Check that the HTMLExporter does not embed images by default"""

        with self.create_temp_cwd(
            ["notebook5_embed_images.ipynb", "containerized_deployments.jpeg"]
        ):
            self.nbconvert("notebook5_embed_images --log-level 0 --to html")
            assert os.path.isfile("notebook5_embed_images.html")
            with open("notebook5_embed_images.html", encoding="utf8") as f:
                text = f.read()
                assert "./containerized_deployments.jpeg" in text
                assert 'src="./containerized_deployments.jpeg"' in text
                assert text.count("data:image/jpeg;base64") == 0

    def test_embedding_images_htmlexporter(self):
        """Check that the HTMLExporter embeds images if needed"""

        with self.create_temp_cwd(
            ["notebook5_embed_images.ipynb", "containerized_deployments.jpeg"]
        ):
            self.nbconvert("notebook5_embed_images --log-level 0 --to html --embed-images")
            assert os.path.isfile("notebook5_embed_images.html")
            with open("notebook5_embed_images.html", encoding="utf8") as f:
                text = f.read()
                assert "./containerized_deployments.jpeg" not in text
                assert "src='./containerized_deployments.jpeg'" not in text
                assert text.count("data:image/jpeg;base64") == 3

    def test_execute_widgets_from_nbconvert(self):
        """Check jupyter widgets render"""
        notebook_name = "Unexecuted_widget"
        with self.create_temp_cwd([f"{notebook_name}.ipynb"]):
            self.nbconvert(f"{notebook_name}.ipynb --execute --log-level 0 --to html")
            assert os.path.isfile(f"{notebook_name}.html")
            with open(f"{notebook_name}.html", encoding="utf8") as f:
                text = f.read()
                assert '<script type="application/vnd.jupyter.widget-view+json">' in text
                assert '<script type="application/vnd.jupyter.widget-state+json">' in text

    def test_execute_multiple_notebooks(self):
        """Check jupyter widgets render in case of batch convert"""
        notebook_name = "Unexecuted_widget"
        with self.create_temp_cwd([f"{notebook_name}*.ipynb"]):
            self.nbconvert("*.ipynb --execute --log-level 0 --to html")

            for name in (notebook_name, f"{notebook_name}_2"):
                assert os.path.isfile(f"{name}.html")
                with open(f"{name}.html", encoding="utf8") as f:
                    text = f.read()
                    assert '<script type="application/vnd.jupyter.widget-view+json">' in text
                    assert '<script type="application/vnd.jupyter.widget-state+json">' in text

    def test_output_base(self):
        """
        Check various configurations of output_base (--output)
        """
        notebook_names = [
            "notebook1",
            "notebook2",
        ]
        with self.create_temp_cwd([x + ".ipynb" for x in notebook_names]):
            self.nbconvert("*.ipynb --output '{notebook_name}_test_addition.asd' --to markdown")

            for nbn in notebook_names:
                assert os.path.isfile(f"{nbn}_test_addition.asd.md")

        with self.create_temp_cwd([x + ".ipynb" for x in notebook_names]):
            self.nbconvert("*.ipynb --to markdown")

            for nbn in notebook_names:
                assert os.path.isfile(f"{nbn}.md")

        # Test single output with static output name
        nbname = notebook_names[0]
        with self.create_temp_cwd([nbname + ".ipynb"]):
            self.nbconvert(f"{nbname}.ipynb --output notebook_test_name --to markdown")
            assert os.path.isfile("notebook_test_name.md")

            self.nbconvert(f"{nbname}.ipynb --to notebook")
            assert os.path.isfile("notebook1.nbconvert.ipynb")

            self.nbconvert(f"{nbname}.ipynb --to notebook --output out.ipynb")
            assert os.path.isfile("out.ipynb")

        # Test double extension fix
        with self.create_temp_cwd([notebook_names[0] + ".ipynb"]):
            self.nbconvert("*.ipynb --output notebook_test_name.md --to markdown")
            assert os.path.isfile("notebook_test_name.md")
            assert not os.path.isfile("notebook_test_name.md.md")

    def test_same_filename_different_dir(self):
        """
        Check if files with same name in different directories pose a problem
        """
        with self.create_temp_cwd() as temp_wd:
            self.copy_files_to(["notebook1.ipynb"], temp_wd + "/dir1")
            self.copy_files_to(["notebook1.ipynb"], temp_wd + "/dir2")
            self.nbconvert("dir1/notebook1.ipynb dir2/notebook1.ipynb --to markdown")
            assert os.path.isfile("dir1/notebook1.md")
            assert os.path.isfile("dir2/notebook1.md")
